import json
import os
import sys
from collections import namedtuple
from string import Template

import requests

t = sys.path[0]
sys.path.remove(t)
from github import Github

sys.path.insert(0, t)

Repo = namedtuple('Repo', ['owner', 'repo'])
Issue = namedtuple('Issue', ['owner', 'repo', 'number'])


class Context:
    payload: dict
    event_name: str
    sha: str
    ref: str
    workflow: str
    action: str
    actor: str
    job: str
    run_number: int
    run_id: int
    api_url: str
    server_url: str
    graphql_url: str

    def __init__(self):
        self.payload = {}
        file = os.getenv('GITHUB_EVENT_PATH')
        if file:
            with open(file, 'r', encoding='utf-8', newline='') as f:
                self.payload = json.load(f)
        self.event_name = os.getenv('GITHUB_EVENT_NAME', '')
        self.sha = os.getenv('GITHUB_SHA', '')
        self.ref = os.getenv('GITHUB_REF', '')
        self.workflow = os.getenv('GITHUB_WORKFLOW', '')
        self.action = os.getenv('GITHUB_ACTION', '')
        self.job = os.getenv('GITHUB_JOB', '')
        self.run_number = int(os.getenv('GITHUB_RUN_NUMBER', '0'))
        self.run_id = int(os.getenv('GITHUB_RUN_ID', '0'))
        self.api_url = os.getenv('GITHUB_API_URL', 'https://api.github.com')
        self.server_url = os.getenv('GITHUB_SERVER_URL', 'https://github.com')
        self.graphql_url = os.getenv('GITHUB_GRAPHQL_URL', 'https://api.github.com/graphql')

    def _repo(self):
        if os.getenv('GITHUB_REPOSITORY'):
            owner, repo = os.getenv('GITHUB_REPOSITORY').split('/')
            return owner, repo
        repository = self.payload.get('repository')
        if not repository:
            return None, None
        return repository['owner']['login'], repository['name']

    def _issue(self):
        payload = self.payload
        actor = payload.get('issue') or payload.get('pull_request') or payload
        owner, repo = self._repo()
        return owner, repo, actor.get('number')

    @property
    def repo(self):
        return Repo(*self._repo())

    @property
    def issue(self):
        return Issue(*self._issue())


class Octokit:
    def __init__(self, token):
        self.token = token
        self.headers = {
            'Authorization': f'Bearer {self.token}'
        }
        self.rest = Github(token)

    def graphql(self, query: str, variables: dict) -> dict:
        q = {
            'query': Template(query).substitute(variables)
        }
        resp = requests.post('https://api.github.com/graphql', json=q, headers=self.headers)
        if resp.status_code != 200:
            raise Exception(f'Query failed to run by returning code of {resp.status_code}. {query}')
        return resp.json()


def get_octokit(token):
    return Octokit(token)



import os
from urllib.parse import quote

import requests


class OidcClient:

    @staticmethod
    def get_request_token() -> str:
        token = os.getenv('ACTIONS_ID_TOKEN_REQUEST_TOKEN')
        if not token:
            raise Exception('Unable to get ACTIONS_ID_TOKEN_REQUEST_TOKEN env variable')
        return token

    @staticmethod
    def get_id_token_url() -> str:
        runtime_url = os.getenv('ACTIONS_ID_TOKEN_REQUEST_URL')
        if not runtime_url:
            raise Exception('Unable to get ACTIONS_ID_TOKEN_REQUEST_URL env variable')
        return runtime_url

    @staticmethod
    def get_call(id_token_url: str):
        headers = {
            'Authorization': 'Bearer ' + OidcClient.get_request_token()
        }
        resp = requests.get(id_token_url, headers=headers)
        if resp.status_code != 200:
            raise Exception(f'Failed to get ID Token. \n'
                            f'Error Code: {resp.status_code}\n'
                            f'Result: {resp.text}')
        res = resp.json()
        if not res.get('value'):
            raise Exception('Response json body do not have ID Token field')
        return res['value']

    @staticmethod
    def get_id_token(audience: str = None):
        try:
            # New ID Token is requested from action service
            id_token_url = OidcClient.get_id_token_url()
            if audience:
                encoded_audience = quote(audience)
                id_token_url = f'{id_token_url}&audience={encoded_audience}'

            id_token = OidcClient.get_call(id_token_url)
            return id_token
        except Exception as e:
            raise Exception(f'Error message: {str(e)}')



import json
import os
from typing import List


class AnnotationProperties:
    def __init__(self, title: str = None, file: str = None,
                 start_line: int = None, end_line: int = None,
                 start_column: int = None, end_column: int = None):
        # A title for the annotation.
        self.title = title

        # The path of the file for which the annotation should be created.
        self.file = file

        # The start line for the annotation.
        self.start_line = start_line

        # The end line for the annotation. Defaults to `start_line` when `start_line` is provided.
        self.end_line = end_line

        # The start column for the annotation. Cannot be sent when `start_line` and `end_line` are different values.
        self.start_column = start_column

        # The start column for the annotation. Cannot be sent when `start_line` and `end_line` are different values.
        # Defaults to `start_column` when `start_column` is provided.
        self.end_column = end_column


class OctokitOptions:
    def __init__(self, auth_strategy=None, auth=None, user_agent: str = None,
                 previews: List[str] = None, base_url: str = None, log=None,
                 request=None, time_zone: str = None):
        self.auth_strategy = auth_strategy
        self.auth = auth
        self.user_agent = user_agent
        self.previews = previews
        self.base_url = base_url
        self.log = log
        self.request = request
        self.time_zone = time_zone


def to_command_value(input_val) -> str:
    """
    Sanitizes an input into a string so it can be passed into issueCommand safely
    :param input_val: input to sanitize into a string
    :return: string
    """
    if input_val is None:
        return ''
    if isinstance(input_val, str):
        return str(input_val)
    return json.dumps(input_val)


def to_command_properties(annotation_properties: AnnotationProperties) -> dict:
    if not annotation_properties:
        return {}
    return {
        'title': annotation_properties.title,
        'file': annotation_properties.file,
        'line': annotation_properties.start_line,
        'endLine': annotation_properties.end_line,
        'col': annotation_properties.start_column,
        'endColumn': annotation_properties.end_column
    }


def get_api_base_url() -> str:
    return os.getenv('GITHUB_API_URL', 'https://api.github.com')


def get_auth_string(token: str, **options) -> str:
    options = OctokitOptions(**options)
    if not token and not options.auth:
        raise Exception('Parameter token or opts.auth is required')
    return options.auth if isinstance(options.auth, str) else f'token {token}'



import asyncio
import errno
import logging
import os
import platform
import re
from functools import wraps, partial
from pathlib import Path
from stat import S_ISDIR

from typing import List

IS_WINDOWS = platform.system() == 'Windows'


def wrap(func):
    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        p = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, p)

    return run


stat = wrap(os.stat)

lstat = wrap(os.lstat)
rename = wrap(os.rename)
remove = wrap(os.remove)
mkdir = wrap(os.mkdir)
rmdir = wrap(os.rmdir)

if hasattr(os, "sendfile"):
    sendfile = wrap(os.sendfile)


async def exists(fs_path: str):
    try:
        await stat(fs_path)
    except IOError as e:
        if e.errno == errno.ENOENT:
            return False
        raise e
    return True


async def is_directory(fs_path: str, use_stat: bool = False):
    stats = await stat(fs_path) if use_stat else await lstat(fs_path)
    return S_ISDIR(stats.st_mode)


async def is_rooted(p: str) -> bool:
    p = normalize_separators(p)
    if not p:
        raise Exception('is_rooted() parameter "p" cannot be empty')

    if IS_WINDOWS:
        return p.startswith('\\') or re.match(r'^[A-Z]', p, re.IGNORECASE) is not None
    return p.startswith('/')


async def mkdir_p(fs_path: str, max_depth: int = 1000, depth: int = 1):
    if not fs_path:
        raise Exception('a path argument must be provided')
    fs_path = Path(fs_path).resolve()
    if depth >= max_depth:
        return mkdir(fs_path)

    try:
        await mkdir(fs_path)
        return
    except IOError as e:
        if e.errno == errno.ENOENT:
            await mkdir_p(str(fs_path), max_depth, depth + 1)
            await mkdir(fs_path)
            return
        try:
            stats = await stat(fs_path)
        except Exception as e2:
            raise e2
        if not S_ISDIR(stats.st_mode):
            raise e


async def try_get_executable_path(file_path: str, extensions: List[str]):
    stats = None
    try:
        stats = await stat(file_path)
    except IOError as e:
        if e.errno != errno.ENOENT:
            logging.log(f'Unexpected error attempting to determine if executable file exists "{file_path}": {e}')
    if stats and Path(file_path).is_file():
        if IS_WINDOWS:
            upper_ext = Path(file_path).suffix.upper()
            for valid_ext in extensions:
                if valid_ext.upper() == upper_ext:
                    return file_path
        else:
            if is_unix_executable(file_path):
                return file_path
    return ''


def normalize_separators(p: str) -> str:
    return os.path.normpath(p or '')


def is_unix_executable(file_path) -> bool:
    return os.access(file_path, os.X_OK)



import json
import os

from actions_toolkit.github import Context, Repo, Issue

os.environ['GITHUB_REPOSITORY'] = 'yanglbme/actions-toolkit'
os.environ['GITHUB_EVENT_PATH'] = os.path.join(os.getcwd(), 'payload.json')
context = Context()

with open('./payload.json', 'r', encoding='utf-8') as f:
    assert context.payload == json.load(f)

os.environ.pop('GITHUB_EVENT_PATH', None)
context = Context()
assert context.payload == {}

assert context.repo == Repo(owner='yanglbme', repo='actions-toolkit')

os.environ.pop('GITHUB_REPOSITORY', None)
context.payload['repository'] = {
    'name': 'test',
    'owner': {
        'login': 'user'
    }
}
assert context.repo == Repo(owner='user', repo='test')

os.environ['GITHUB_REPOSITORY'] = 'yanglbme/actions-toolkit'
os.environ['GITHUB_EVENT_PATH'] = os.path.join(os.getcwd(), 'payload.json')
context = Context()
assert context.issue == Issue(owner='yanglbme', repo='actions-toolkit', number=1)

os.environ.pop('GITHUB_REPOSITORY', None)
context.payload = {
    'pull_request': {
        'number': 2
    },
    'repository': {
        'owner': {
            'login': 'user'
        },
        'name': 'test'
    }
}
assert context.issue == Issue(owner='user', repo='test', number=2)

os.environ.pop('GITHUB_REPOSITORY', None)
context.payload = {
    'number': 2,
    'repository': {
        'owner': {
            'login': 'user'
        },
        'name': 'test'
    }
}
assert context.issue == Issue(owner='user', repo='test', number=2)



import asyncio
import io
import os
import sys

from actions_toolkit import core
from actions_toolkit.utils import to_command_properties, AnnotationProperties

test_env_vars = {
    'my var': '',
    'special char var \r\n];': '',
    'my var2': '',
    'my secret': '',
    'special char secret \r\n];': '',
    'my secret2': '',
    'PATH': f'path1{os.pathsep}path2',

    # Set inputs
    'INPUT_MY_INPUT': 'val',
    'INPUT_MISSING': '',
    'INPUT_SPECIAL_CHARS_\'\t"\\': '\'\t"\\ response ',
    'INPUT_MULTIPLE_SPACES_VARIABLE': 'I have multiple spaces',
    'INPUT_BOOLEAN_INPUT': 'true',
    'INPUT_BOOLEAN_INPUT_TRUE1': 'true',
    'INPUT_BOOLEAN_INPUT_TRUE2': 'True',
    'INPUT_BOOLEAN_INPUT_TRUE3': 'TRUE',
    'INPUT_BOOLEAN_INPUT_FALSE1': 'false',
    'INPUT_BOOLEAN_INPUT_FALSE2': 'False',
    'INPUT_BOOLEAN_INPUT_FALSE3': 'FALSE',
    'INPUT_WRONG_BOOLEAN_INPUT': 'wrong',
    'INPUT_WITH_TRAILING_WHITESPACE': '  some val  ',

    'INPUT_MY_INPUT_LIST': 'val1\nval2\nval3',

    # Save inputs
    'STATE_TEST_1': 'state_val',

    # File Commands
    'GITHUB_PATH': '',
    'GITHUB_ENV': ''
}

for k, v in test_env_vars.items():
    os.environ[k] = v

file_path = os.path.join(os.getcwd(), 'test')
if not os.path.isdir(file_path):
    os.makedirs(file_path)


def call(func, *args, **kw):
    output = io.StringIO()
    sys.stdout = output
    func(*args, **kw)
    sys.stdout = sys.__stdout__
    return output.getvalue()


def create_file_command_file(command: str):
    path = os.path.join(file_path, command)
    os.environ[f'GITHUB_{command}'] = path
    with open(path, 'a', encoding='utf-8', newline='') as fs:
        fs.write('')


def verify_file_command(command: str, expected_contents: str):
    path = os.path.join(file_path, command)
    with open(path, 'r', encoding='utf-8', newline='') as fs:
        contents = fs.read()
        assert contents == expected_contents
    os.unlink(path)
    os.environ.pop(f'GITHUB_{command}', None)


assert call(core.export_variable, 'my var', 'var val') == f'::set-env name=my var::var val{os.linesep}'

assert call(core.export_variable, 'special char var \r\n,:',
            'special val') == f'::set-env name=special char var %0D%0A%2C%3A::special val{os.linesep}'
assert os.getenv('special char var \r\n,:') == 'special val'

assert call(core.export_variable, 'my var2', 'var val\r\n') == f'::set-env name=my var2::var val%0D%0A{os.linesep}'
assert os.getenv('my var2') == 'var val\r\n'

assert call(core.export_variable, 'my var', True) == f'::set-env name=my var::true{os.linesep}'

assert call(core.export_variable, 'my var', 5) == f'::set-env name=my var::5{os.linesep}'

command = 'ENV'
create_file_command_file(command)
core.export_variable('my var', 'var val')
verify_file_command(command,
                    f'my var<<_GitHubActionsFileCommandDelimeter_{os.linesep}var val'
                    f'{os.linesep}_GitHubActionsFileCommandDelimeter_{os.linesep}')

command = 'ENV'
create_file_command_file(command)
core.export_variable('my var', True)
verify_file_command(command,
                    f'my var<<_GitHubActionsFileCommandDelimeter_{os.linesep}true'
                    f'{os.linesep}_GitHubActionsFileCommandDelimeter_{os.linesep}')

command = 'ENV'
create_file_command_file(command)
core.export_variable('my var', 5)
verify_file_command(command,
                    f'my var<<_GitHubActionsFileCommandDelimeter_{os.linesep}5'
                    f'{os.linesep}_GitHubActionsFileCommandDelimeter_{os.linesep}')

assert call(core.set_secret, 'secret val') == f'::add-mask::secret val{os.linesep}'

command = 'PATH'
create_file_command_file(command)
core.add_path('myPath')
assert os.getenv('PATH') == f'myPath{os.pathsep}path1{os.pathsep}path2'
verify_file_command(command, f'myPath{os.linesep}')

assert call(core.add_path, 'myPath') == f'::add-path::myPath{os.linesep}'

assert core.get_input('my input') == 'val'

assert core.get_input('my input', required=True) == 'val'

try:
    core.get_input('missing', required=True)
except Exception as e:
    assert str(e) == 'Input required and not supplied: MISSING'
else:
    raise Exception('Expected raise Exception but it did not')

assert core.get_input('missing', required=False) == ''

assert core.get_input('My InPuT') == 'val'

assert core.get_input('special chars_\'\t"\\') == '\'\t"\\ response'

assert core.get_input('multiple spaces variable') == 'I have multiple spaces'

assert core.get_multiline_input('my input list') == ['val1', 'val2', 'val3']

assert core.get_input('with trailing whitespace') == 'some val'

assert core.get_input('with trailing whitespace', trim_whitespace=True) == 'some val'

assert core.get_input('with trailing whitespace', trim_whitespace=False) == '  some val  '

assert core.get_boolean_input('boolean input') is True

assert core.get_boolean_input('boolean input', required=True) is True

assert core.get_boolean_input('boolean input true1') is True
assert core.get_boolean_input('boolean input true2') is True
assert core.get_boolean_input('boolean input true3') is True
assert core.get_boolean_input('boolean input false1') is False
assert core.get_boolean_input('boolean input false2') is False
assert core.get_boolean_input('boolean input false3') is False

try:
    core.get_boolean_input('wrong boolean input')
except Exception as e:
    assert str(e) == 'Input does not meet YAML 1.2 "Core Schema" specification: wrong boolean input\n' \
                     'Support boolean input list: `true | True | TRUE | false | False | FALSE`'
else:
    raise Exception('Expected raise Exception but it did not')

assert call(core.set_output, 'some output', 'some value') == \
       f'{os.linesep}::set-output name=some output::some value{os.linesep}'

assert call(core.set_output, 'some output', False) == f'{os.linesep}::set-output name=some output::false{os.linesep}'

assert call(core.set_output, 'some output', 1.01) == f'{os.linesep}::set-output name=some output::1.01{os.linesep}'

assert call(core.error, 'Error message') == f'::error::Error message{os.linesep}'

assert call(core.error, 'Error message\r\n\n') == f'::error::Error message%0D%0A%0A{os.linesep}'

message = 'this is my error message'
error = Exception(message)
assert call(core.error, error) == f'::error::Error: {message}{os.linesep}'

assert call(core.error, error, title='A title', file='root/test.txt',
            start_column=1, end_column=2, start_line=5, end_line=5) == \
       f'::error title=A title,file=root/test.txt,line=5,endLine=5,col=1,endColumn=2::Error: {message}{os.linesep}'

assert call(core.warning, 'Warning') == f'::warning::Warning{os.linesep}'

assert call(core.warning, '\r\nwarning\n') == f'::warning::%0D%0Awarning%0A{os.linesep}'

assert call(core.warning, error) == f'::warning::Error: {message}{os.linesep}'

assert call(core.warning, error, title='A title', file='root/test.txt',
            start_column=1, end_column=2, start_line=5, end_line=5) == \
       f'::warning title=A title,file=root/test.txt,line=5,endLine=5,col=1,endColumn=2::Error: {message}{os.linesep}'

assert call(core.notice, '\r\nnotice\n') == f'::notice::%0D%0Anotice%0A{os.linesep}'

assert call(core.notice, error) == f'::notice::Error: {message}{os.linesep}'

assert call(core.notice, error, title='A title', file='root/test.txt',
            start_column=1, end_column=2, start_line=5, end_line=5) == \
       f'::notice title=A title,file=root/test.txt,line=5,endLine=5,col=1,endColumn=2::Error: {message}{os.linesep}'

annotation_properties = AnnotationProperties(title='A title', file='root/test.txt', start_column=1, end_column=2,
                                             start_line=5, end_line=5)
command_properties = to_command_properties(annotation_properties)
assert command_properties['title'] == 'A title'
assert command_properties['file'] == 'root/test.txt'
assert command_properties['col'] == 1
assert command_properties['endColumn'] == 2
assert command_properties['line'] == 5
assert command_properties['endLine'] == 5
assert command_properties.get('startColumn') is None
assert command_properties.get('startLine') is None

assert call(core.start_group, 'my-group') == f'::group::my-group{os.linesep}'

assert call(core.end_group) == f'::endgroup::{os.linesep}'


async def f():
    async def in_group():
        sys.stdout.write('in my group\n')
        return True

    result = await core.group('mygroup', in_group)
    assert result is True


assert call(asyncio.run, f()) == f'::group::mygroup{os.linesep}in my group\n::endgroup::{os.linesep}'

assert call(core.debug, 'Debug') == f'::debug::Debug{os.linesep}'

assert call(core.debug, '\r\ndebug\n') == f'::debug::%0D%0Adebug%0A{os.linesep}'

assert call(core.save_state, 'state_1', 'some value') == f'::save-state name=state_1::some value{os.linesep}'

assert call(core.save_state, 'state_1', 1) == f'::save-state name=state_1::1{os.linesep}'

assert call(core.save_state, 'state_1', True) == f'::save-state name=state_1::true{os.linesep}'

assert core.get_state('TEST_1') == 'state_val'

os.environ.pop('RUNNER_DEBUG', None)
assert core.is_debug() is False
os.environ['RUNNER_DEBUG'] = '1'
assert core.is_debug() is True
os.environ.pop('RUNNER_DEBUG', None)

assert call(core.set_command_echo, True) == f'::echo::on{os.linesep}'

assert call(core.set_command_echo, False) == f'::echo::off{os.linesep}'



import enum
import os
import sys
from typing import List, Union

from actions_toolkit.command import issue_command, issue
from actions_toolkit.file_command import issue_command as issue_file_command
from actions_toolkit.oidc_utils import OidcClient
from actions_toolkit.utils import to_command_value, to_command_properties, AnnotationProperties


class InputOptions:
    """Interface for getInput options"""

    def __init__(self, required: bool = False, trim_whitespace: bool = True):
        """Optional. Whether the input is required. If required and not present, will throw. Defaults to false"""
        self.required = required

        """Optional. Whether leading/trailing whitespace will be trimmed for the input. Defaults to true"""
        self.trim_whitespace = trim_whitespace


@enum.unique
class ExitCode(enum.IntEnum):
    """The code to exit an action"""

    """A code indicating that the action was successful"""
    success = 0

    """A code indicating that the action was a failure"""
    failure = 1


# -----------------------------------------------------------------------
# Variables
# -----------------------------------------------------------------------

def export_variable(name: str, val):
    """
    Sets env variable for this action and future actions_toolkit in the job
    :param name: the name of the variable to set
    :param val: the value of the variable. Non-string values will be converted to a string via json.dumps
    :return: void
    """
    converted_val = to_command_value(val)
    os.environ[name] = converted_val

    file_path = os.getenv('GITHUB_ENV')
    if file_path:
        delimiter = '_GitHubActionsFileCommandDelimeter_'
        command_value = f'{name}<<{delimiter}{os.linesep}{converted_val}{os.linesep}{delimiter}'
        issue_file_command('ENV', command_value)
    else:
        issue_command('set-env', dict(name=name), converted_val)


def set_secret(secret: str):
    """
    Registers a secret which will get masked from logs
    :param secret: value of the secret
    :return: void
    """
    issue_command('add-mask', {}, secret)


def add_path(input_path: str):
    """
    Prepends inputPath to the PATH (for this action and future actions_toolkit)
    :param input_path: value of the path
    :return: void
    """
    file_path = os.getenv('GITHUB_PATH')
    if file_path:
        issue_file_command('PATH', input_path)
    else:
        issue_command('add-path', {}, input_path)
    os.environ['PATH'] = f'{input_path}{os.pathsep}{os.getenv("PATH")}'


def get_input(name: str, **options) -> str:
    """
    Gets the value of an input.
    Unless trimWhitespace is set to false in InputOptions, the value is also trimmed.
    Returns an empty string if the value is not defined.
    """
    options = InputOptions(**options)
    name = name.replace(' ', '_').upper()
    val = os.getenv(f'INPUT_{name}', '')
    if options.required and not val:
        raise Exception(f'Input required and not supplied: {name}')
    if not options.trim_whitespace:
        return val
    return val.strip()


def get_multiline_input(name: str, **options) -> List[str]:
    """
    Gets the values of an multiline input.  Each value is also trimmed.
    """
    return list(filter(lambda x: x != '', get_input(name, **options).split("\n")))


def get_boolean_input(name: str, **options) -> bool:
    """
    Gets the input value of the boolean type in the YAML 1.2 "core schema" specification.
    Support boolean input list: `true | True | TRUE | false | False | FALSE` .
    The return value is also in boolean type.
    ref: https://yaml.org/spec/1.2/spec.html#id2804923
    """
    true_value = ['true', 'True', 'TRUE']
    false_value = ['false', 'False', 'FALSE']
    val = get_input(name, **options)
    if val in true_value:
        return True
    if val in false_value:
        return False
    raise TypeError(f'Input does not meet YAML 1.2 "Core Schema" specification: {name}\n'
                    f'Support boolean input list: `true | True | TRUE | false | False | FALSE`')


def set_output(name: str, value):
    """
    Sets the value of an output.
    :param name: name of the output to set
    :param value: value to store. Non-string values will be converted to a string via json.dumps
    :return: void
    """
    sys.stdout.write(os.linesep)
    issue_command('set-output', dict(name=name), value)


def set_command_echo(enabled: bool):
    """
    Enables or disables the echoing of commands into stdout for the rest of the step.
    Echoing is disabled by default if ACTIONS_STEP_DEBUG is not set.
    """
    issue('echo', 'on' if enabled else 'off')


# -----------------------------------------------------------------------
# Results
# -----------------------------------------------------------------------

def set_failed(message: Union[str, Exception]):
    """
    Sets the action status to failed.
    When the action exits it will be with an exit code of 1
    :param message: add error issue message
    :return: void
    """
    error(message)
    sys.exit(ExitCode.failure.value)


def is_debug() -> bool:
    """Gets whether Actions Step Debug is on or not"""
    return os.getenv('RUNNER_DEBUG') == '1'


def debug(message: str):
    """
    Writes debug message to user log
    :param message: debug message
    :return: void
    """
    issue_command('debug', {}, message)


def error(message: Union[str, Exception], **properties):
    """
    Adds an error issue
    :param message: error issue message. Errors will be converted to string via str()
    """
    properties = AnnotationProperties(**properties) if properties else properties
    if isinstance(message, Exception):
        message = f'Error: {str(message)}'
    issue_command('error', to_command_properties(properties), message)


def warning(message: Union[str, Exception], **properties):
    """
    Adds a warning issue
    :param message: warning issue message. Errors will be converted to string via str()
    """
    properties = AnnotationProperties(**properties) if properties else properties
    if isinstance(message, Exception):
        message = f'Error: {str(message)}'
    issue_command('warning', to_command_properties(properties), message)


def notice(message: Union[str, Exception], **properties):
    """
    Adds a notice issue
    :param message: notice issue message. Errors will be converted to string via str()
    """
    properties = AnnotationProperties(**properties) if properties else properties
    if isinstance(message, Exception):
        message = f'Error: {str(message)}'
    issue_command('notice', to_command_properties(properties), message)


def info(message: str):
    """
    Writes info to log with console.log.
    :param message: info message
    :return: void
    """
    sys.stdout.write(message + os.linesep)


def start_group(name: str):
    """
    Begin an output group.

    Output until the next `groupEnd` will be foldable in this group
    :param name: The name of the output group
    :return: void
    """
    issue('group', name)


def end_group():
    """End an output group."""
    issue('endgroup')


async def group(name: str, fn):
    """
    Wrap an asynchronous function call in a group.

    Returns the same type as the function itself.
    :param name: The name of the group
    :param fn: The function to wrap in the group
    """
    start_group(name)
    try:
        result = await fn()
    finally:
        end_group()
    return result


# -----------------------------------------------------------------------
# Wrapper action state
# -----------------------------------------------------------------------

def save_state(name: str, value):
    """
    Saves state for current action, the state can only be retrieved by this action's post job execution.
    :param name: name of the state to store
    :param value: value to store. Non-string values will be converted to a string via json.dumps
    :return: void
    """
    issue_command('save-state', dict(name=name), value)


def get_state(name: str) -> str:
    """
    Gets the value of an state set by this action's main execution.
    :param name: name of the state to get
    :return: string
    """
    return os.getenv(f'STATE_{name}', '')


def get_id_token(aud: str = None) -> str:
    id_token = OidcClient.get_id_token(aud)
    set_secret(id_token)
    return id_token
